from ..interface import Problem, ProblemConfig
from .utils import create_dataset_directory_tree
import os
import torch
import numpy as np
from fenics import *
from itertools import product
from .utils import create_dataset_directory_tree, midpoints, get_param_space

set_log_active(False)

class Heat2D(Problem):
    '''
    2D Heat Equation:

    DuDt = alpha*Laplacian(u(x))        x in [-1,1]^2 x (0,T)
    u(x,t) = uD(x,t)                    x on D[-1,1]^2
    u(x,0) = exp(-(x^2+y^2)/2)          x in [-1,1]^2 x {0} 

    '''
    def __init__(self, config: ProblemConfig):
        super().__init__(config)

        self.directory = config.directory
        create_dataset_directory_tree(self.directory)
        self.save_vtk = config.save_vtk

        self.set_domain(config)

        self.parameters = config.parameters
        
        self.param_instances = get_param_space(self.parameters[0], config)

        self.Nh = self.mesh.num_vertices()    # number of DOF
        self.Nt = config.Nt                   # number of timesteps              
        self.Np = 2                           # num param + time
        self.N = len(self.param_instances)

        self.time_interval = config.time_interval   # final time
        self.dt = (self.time_interval[1]-self.time_interval[0]) / self.Nt 

        self.S = torch.empty(size=(self.N,self.Nt,self.Nh),dtype=torch.float32)
        self.P = torch.empty(size=(self.N,self.Nt,self.Np),dtype=torch.float32)


    def set_domain(self,config):
        if config.domain == 'square':
            self.n = config.n
            self.mesh = RectangleMesh(Point(-1, -1), Point(1, 1), self.n, self.n)
            self.V = FunctionSpace(self.mesh, 'P', 1)
            self.bc = DirichletBC(self.V, Constant(0), 'on_boundary')
            File(os.path.join(self.directory,'mesh/mesh.pvd')) << self.mesh
        else:
            raise SyntaxError('only square is available as domain type')


    def solve(self):
        for i,p1 in enumerate(self.param_instances):
            print('Solving for parameters instance #',i,' :', p1)

            if self.save_vtk:
                vtkfile = File(os.path.join(self.directory,'vtk','solution_'+str(i),'solution.pvd'))

            # Define initial value
            u_0 = Expression('exp(-a*(pow(x[0], 2) + pow(x[1], 2)))', degree=2, a=5)
            u_n = interpolate(u_0, self.V)

            # Define variational problem
            u = TrialFunction(self.V)
            v = TestFunction(self.V)
            f = Constant(0)

            lhs = u*v*dx + p1*self.dt*dot(grad(u), grad(v))*dx 
            rhs = (u_n + self.dt*f)*v*dx

            # Time-stepping0
            u_sol = Function(self.V)
            t = self.time_interval[0]

            for j in range(self.Nt):

                t += self.dt
                solve(lhs == rhs, u_sol, self.bc)

                self.S[i,j,:] = torch.from_numpy(u_sol.vector().get_local(vertex_to_dof_map(self.V)))
                self.P[i,j,:] = torch.tensor([p1,t])
                
                if self.save_vtk:
                    vtkfile << (u_sol, t)

                u_n.assign(u_sol)

    def save_dataset(self):
        torch.save(self.S, os.path.join(self.directory,'tensors/snapshots/S.pt'))
        torch.save(self.P, os.path.join(self.directory,'tensors/parameters/P.pt'))
        print('Dataset correctly saved to:', self.directory)
        


import os
import numpy as np

def create_dataset_directory_tree(path: str):

    mesh_path = os.path.join(path,"mesh")
    tensors_path = os.path.join(path,"tensors")
    snapshots_path = os.path.join(tensors_path,"snapshots")
    params_path = os.path.join(tensors_path,"parameters")
    vtk_path = os.path.join(path,"vtk")

    makedir(path)
    makedir(mesh_path)
    makedir(tensors_path)
    makedir(snapshots_path)
    makedir(params_path)
    makedir(vtk_path)
    
    print('Dataset directory tree created')


def makedir(path):
    if not os.path.isdir(path):
        os.makedirs(path)


def midpoints(array: np.array):
    return (array[:-1] + array[1:])/2

def get_param_space(p, config):
    if len(p)==1:
        return p
    elif len(p)==3:
        p_space = np.linspace(p[0],p[1],p[2])
        if config.midpoints:
            p_space = midpoints(p_space)
        return p_space.tolist()
            
    else:
        raise ValueError('Input parameters must be a single valued array or a triple.')



from ..interface import Problem, ProblemConfig
from .utils import create_dataset_directory_tree
import os
import torch
import numpy as np
from fenics import *
from itertools import product
from .utils import create_dataset_directory_tree, midpoints, get_param_space

set_log_active(False)

class ADR2D(Problem):
    '''
    2D Advection-Diffusion-Reaction Equation:

    ...

    '''
    def __init__(self, config: ProblemConfig):
        super().__init__(config)

        self.directory = config.directory
        create_dataset_directory_tree(self.directory)
        self.save_vtk = config.save_vtk

        self.set_domain(config)

        self.parameters = config.parameters
        
        self.p1_space = get_param_space(self.parameters[0], config)
        self.p2_space = get_param_space(self.parameters[1], config)
        self.p3_space = get_param_space(self.parameters[2], config)
        self.p4_space = get_param_space(self.parameters[3], config)
        self.param_space = [self.p1_space, self.p2_space, self.p3_space, self.p4_space]
        self.param_instances = product(*self.param_space)

        self.Nh = self.mesh.num_vertices()    # number of DOF
        self.Nt = config.Nt                   # number of timesteps              
        self.Np = 5                           # num param + time
        self.N = len(sum(self.param_space,[]))

        self.time_interval = config.time_interval   # final time
        self.dt = (self.time_interval[1]-self.time_interval[0]) / self.Nt 

        self.S = torch.empty(size=(self.N,self.Nt,self.Nh),dtype=torch.float32)
        self.P = torch.empty(size=(self.N,self.Nt,self.Np),dtype=torch.float32)


    def set_domain(self,config):
        if config.domain == 'square':
            self.n = config.n
            self.mesh = RectangleMesh(Point(0, 0), Point(1, 1), self.n, self.n)
            self.V = FunctionSpace(self.mesh, 'P', 1)
            self.bc = DirichletBC(self.V, Constant(0), 'on_boundary')
            File(os.path.join(self.directory,'mesh/mesh.pvd')) << self.mesh
        else:
            raise SyntaxError('only square is available as domain type')


    def solve(self):
        i=0
        for p1,p2,p3,p4 in self.param_instances:
            print('Solving for parameters instance #',i,' :', [p1,p2,p3,p4])

            if self.save_vtk:
                vtkfile = File(os.path.join(self.directory,'vtk','solution_'+str(i),'solution.pvd'))

            # Define initial value
            u_0 = Constant(0.)
            u_n = interpolate(u_0, self.V)

            # Define variational problem
            u = TrialFunction(self.V)
            v = TestFunction(self.V)
            f = Expression("10.*exp(-(pow(x[0]-p3,2)+pow(x[1]-p4,2))/0.0049)", degree=2, p3=p3, p4=p4)
            b = Expression(("cos(2*pi*t/p2)","sin(2*pi*t/p2)"), degree=2, p2=p2, t=0)
            c = Constant(1.)
            a = u*v*dx + p1*self.dt*inner(grad(u),grad(v))*dx + self.dt*inner(b,grad(u))*v*dx + c*self.dt*u*v*dx
            L = u_n*v*dx + self.dt*f*v*dx

            # Time-stepping0
            u_sol = Function(self.V)
            t = self.time_interval[0]

            for j in range(self.Nt):

                t += self.dt
                solve(a == L, u_sol)

                self.S[i,j,:] = torch.from_numpy(u_sol.vector().get_local(vertex_to_dof_map(self.V)))
                self.P[i,j,:] = torch.tensor([p1,p2,p3,p4,t])
                
                if self.save_vtk:
                    vtkfile << (u_sol, t)

                u_n.assign(u_sol)
            i+=1
            
    def save_dataset(self):
        torch.save(self.S, os.path.join(self.directory,'tensors/snapshots/S.pt'))
        torch.save(self.P, os.path.join(self.directory,'tensors/parameters/P.pt'))
        print('Dataset correctly saved to:', self.directory)
        


from .interface import ProblemConfig, parse_config
from .problems import *
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('--config', type=str, default=None, help='Path to the yaml config file')
   
PROBLEMS_DICT = {
    'burgers1d': Burgers1D,
    'heat2d': Heat2D,
    'adr2d': ADR2D,
    'navierstokes2d': NavierStokes2DCylinder
    }

def generate(config: ProblemConfig or str):
    
    if type(config) is str:
        config = parse_config(config)

    problem = PROBLEMS_DICT[config.problem](config)
    problem.solve()
    problem.save_dataset()


if __name__ == "__main__":
    args = parser.parse_args()
    generate(args.config)


import setuptools


setuptools.setup(
    name="pdegen",
    version="0.0.2",
    author="Nicola Farenga",
    author_email="nicola.farenga@mail.polimi.it",
    description="Partial Differential Equations Dataset Generation Library",
    url="https://github.com/farenga/pdegen",
    packages=setuptools.find_packages(),
    install_requires=['torch>=1.5.0', 'numpy>=1.21.0'],
    python_requires='~=3.7',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ]
)


from ..interface import Problem, ProblemConfig
from .utils import create_dataset_directory_tree
import os
import torch
import numpy as np
from fenics import *
from mshr import *
import numpy as np
from itertools import product
from .utils import create_dataset_directory_tree, midpoints, get_param_space

set_log_active(False)

class NavierStokes2DCylinder(Problem):
    '''
    2D Navier-Stokes past a cylinder

    ...

    '''
    def __init__(self, config: ProblemConfig):
        super().__init__(config)

        self.directory = config.directory
        create_dataset_directory_tree(self.directory)
        self.save_vtk = config.save_vtk

        self.set_domain(config)

        self.parameters = config.parameters
        
        self.p1_space = get_param_space(self.parameters[0], config) #   mu: dynamic viscosity 
        self.p2_space = get_param_space(self.parameters[1], config) #   rho: density
        self.param_space = [self.p1_space, self.p2_space]
        self.param_instances = product(*self.param_space)

        self.Nh = self.mesh.num_vertices()    # number of DOF
        print(self.Nh)
        self.Nt = config.Nt                   # number of timesteps              
        self.Np = 3                           # num param + time
        self.N = len(sum(self.param_space,[]))

        self.time_interval = config.time_interval   
        self.dt = (self.time_interval[1]-self.time_interval[0]) / self.Nt 

        self.S = torch.empty(size=(self.N,self.Nt,self.Nh),dtype=torch.float32)
        self.P = torch.empty(size=(self.N,self.Nt,self.Np),dtype=torch.float32)


    def set_domain(self,config):
        if config.domain == 'rectangle':
            self.n = config.n
            channel = Rectangle(Point(0, 0), Point(2.2, 0.41))
            cylinder = Circle(Point(0.2, 0.2), 0.05)
            domain = channel - cylinder
            self.mesh = generate_mesh(domain, self.n)

            self.V = VectorFunctionSpace(self.mesh, 'P', 2)
            self.Q = FunctionSpace(self.mesh, 'P', 1)

            inflow   = 'near(x[0], 0)'
            outflow  = 'near(x[0], 2.2)'
            walls    = 'near(x[1], 0) || near(x[1], 0.41)'
            cylinder = 'on_boundary && x[0]>0.1 && x[0]<0.3 && x[1]>0.1 && x[1]<0.3'

            # Define inflow profile
            inflow_profile = ('4.0*1.5*x[1]*(0.41 - x[1]) / pow(0.41, 2)', '0')

            # Define boundary conditions
            bcu_inflow = DirichletBC(self.V, Expression(inflow_profile, degree=2), inflow)
            bcu_walls = DirichletBC(self.V, Constant((0, 0)), walls)
            bcu_cylinder = DirichletBC(self.V, Constant((0, 0)), cylinder)
            bcp_outflow = DirichletBC(self.Q, Constant(0), outflow)
            self.bcu = [bcu_inflow, bcu_walls, bcu_cylinder]
            self.bcp = [bcp_outflow]

            File(os.path.join(self.directory,'mesh/mesh.pvd')) << self.mesh
        else:
            raise SyntaxError('only rectangle is available as domain type')


    def solve(self):
        i=0
        for p1,p2 in self.param_instances:
            print('Solving for parameters instance #',i,' :', [p1,p2])

            if self.save_vtk:
                vtkfile = File(os.path.join(self.directory,'vtk','solution_'+str(i),'solution.pvd'))

            # Define trial and test functions
            u = TrialFunction(self.V)
            v = TestFunction(self.V)
            p = TrialFunction(self.Q)
            q = TestFunction(self.Q)

            # Define functions for solutions at previous and current time steps
            u_n = Function(self.V)
            u_  = Function(self.V)
            p_n = Function(self.Q)
            p_  = Function(self.Q)

            U  = 0.5*(u_n + u)
            n  = FacetNormal(self.mesh)
            f  = Constant((0, 0))
            k  = Constant(self.dt)
            mu = Constant(p1)
            rho = Constant(p2)

            # Define symmetric gradient
            def epsilon(u):
                return sym(nabla_grad(u))

            # Define stress tensor
            def sigma(u, p):
                return 2*mu*epsilon(u) - p*Identity(len(u))

            # Define variational problem for step 1
            F1 = rho*dot((u - u_n) / k, v)*dx + rho*dot(dot(u_n, nabla_grad(u_n)), v)*dx + inner(sigma(U, p_n), epsilon(v))*dx + dot(p_n*n, v)*ds - dot(mu*nabla_grad(U)*n, v)*ds - dot(f, v)*dx
            a1 = lhs(F1)
            L1 = rhs(F1)

            # Define variational problem for step 2
            a2 = dot(nabla_grad(p), nabla_grad(q))*dx
            L2 = dot(nabla_grad(p_n), nabla_grad(q))*dx - (1/k)*div(u_)*q*dx

            # Define variational problem for step 3
            a3 = dot(u, v)*dx
            L3 = dot(u_, v)*dx - k*dot(nabla_grad(p_ - p_n), v)*dx

            # Assemble matrices
            A1 = assemble(a1)
            A2 = assemble(a2)
            A3 = assemble(a3)

            # Apply boundary conditions to matrices
            [bc.apply(A1) for bc in self.bcu]
            [bc.apply(A2) for bc in self.bcp]

            # Time-stepping
            t = self.time_interval[0]

            for j in range(self.Nt):

                t += self.dt
                print('\tsolving for t=',t)
                # Step 1: Tentative velocity step
                b1 = assemble(L1)
                [bc.apply(b1) for bc in self.bcu]
                solve(A1, u_.vector(), b1, 'bicgstab', 'hypre_amg')

                # Step 2: Pressure correction step
                b2 = assemble(L2)
                [bc.apply(b2) for bc in self.bcp]
                solve(A2, p_.vector(), b2, 'bicgstab', 'hypre_amg')

                # Step 3: Velocity correction step
                b3 = assemble(L3)
                solve(A3, u_.vector(), b3, 'cg', 'sor')

                # ********** here you must stack sol vectors
                #self.S[i,j,:] = torch.from_numpy(u_.vector().get_local(vertex_to_dof_map(self.V)))
                #self.P[i,j,:] = torch.tensor([p1,p2,t])
                
                if self.save_vtk:
                    vtkfile << (u_, t)

                u_n.assign(u_)
                p_n.assign(p_)
            i+=1
            
    def save_dataset(self):
        torch.save(self.S, os.path.join(self.directory,'tensors/snapshots/S.pt'))
        torch.save(self.P, os.path.join(self.directory,'tensors/parameters/P.pt'))
        print('Dataset correctly saved to:', self.directory)
        


from dataclasses import dataclass
import yaml

@dataclass
class ProblemConfig:

    problem: str
    domain: str
    parameters: int or list         # Parameters [a,b,...] or [[a0,a1,Na],[b0,b1,Nb]]
    midpoints: bool = False         # Take the midpoints of the specified parameters spaces
    n: int = 32                     # Number of lateral dof
    Nt: int = None                  # Number of timesteps
    time_interval: list = None      # Time interval [t0,t1]
    directory: str = 'dataset'      # Working directory
    save_vtk: bool = False          # Save vtk visualization files

class Problem:
    def __init__(self, config: ProblemConfig or str):
        pass
    
    def solve(self):
        pass
    
    def save(self):
        pass

def parse_config(filepath):
    
    with open(filepath, 'r') as stream:
        config_file = yaml.safe_load(stream)

    problem_config = ProblemConfig(
        problem = config_file['problem'],
        domain = config_file['domain'],
        parameters = config_file['parameters'],
        midpoints= config_file['midpoints'],
        n = config_file['n'],
        Nt = config_file['Nt'],
        time_interval=config_file['time_interval'],
        directory = config_file['directory'],
        save_vtk = config_file['save_vtk']
    )

    return problem_config
            
